(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(* Run a test and collect statistics *)
open Misc
open Globals
open MiscParser
open Printf
open MySys

(* Check dontrun file *)
let open_norun name chan =
  if is_out () then fprintf chan "if [ ! -f %s.no ]; then\n" name
and close_norun chan =
  if is_out () then output_line chan "fi"

(* Text quotation for cross-compilation *)
let open_quote chan =
  if is_out() then output_line chan "cat <<'EOF'"
and close_quote chan =
  if is_out() then output_line chan "EOF"


let target_os arch =
  if is_out() then
    get_target_os arch
  else begin
    if Sys.file_exists "/mach_kernel"
    then begin 
      Warn.warn "OS: Darwin" ;
      Mac
    end else begin
      Warn.warn "OS: Linux" ;
      Linux
    end
  end

let get_gcc_opts arch =
  let std_opts = "-Wall -std=gnu99 " ^  Globals.get_gcc_opts arch in
  let opts =
    match target_os arch with
    | Mac -> begin match get_word arch with
      | W64 -> std_opts ^ " -m64"
      | W32 -> std_opts ^ " -m32"
      | WXX -> std_opts
    end
    | Linux -> begin match get_word arch with
      | W64 -> std_opts ^ " -m64 -pthread"
      | W32 -> std_opts ^ " -m32 -pthread"
      | WXX -> std_opts ^ " -pthread"
    end
    | AIX -> begin match get_word arch with
      | W64 -> std_opts ^ " -maix64 -pthread"
      | W32 -> std_opts ^ " -maix32 -pthread"
      | WXX -> std_opts ^ " -pthread"
    end in
  let opts = match Globals.get_affinity () with
  | AffinityNone -> opts
  | AffinityIncr _|AffinityRandom|AffinityCustom -> "-D_GNU_SOURCE " ^ opts in
  opts

let get_link_opts arch =
  let def_opts =
    match target_os arch with
    | Mac -> ""
    | Linux|AIX ->
        if  get_clocktime () || get_profiling () then " -lrt"
        else "" in
  let add_opts = Globals.get_link_opt () in
  match def_opts,add_opts with
  | (s,"")|("",s) -> s
  | _,_ -> def_opts ^ " " ^ add_opts

let pp_barrier_loc arch =
  let b = pp_barrier arch in
  match get_barrier arch with
  | TimeBase -> sprintf "%s (%i)" b (get_delay arch)
  | _ -> b

let report_parameters arch chan =
  open_quote chan ;
  fprintf chan "Revision %s, version %s\n" Config.rev Config.version;
  fprintf chan "Command line:%s\n"
    (String.concat ""
       (List.map (sprintf " %s") (Array.to_list Sys.argv))) ;
  output_line chan "Parameters" ;
  dump_bench arch chan ;
  fprintf chan "/* gcc options: %s */\n" (get_gcc_opts arch) ;
  begin match get_link_opt () with
  | "" -> ()
  | s ->   fprintf chan "/* gcc link options: %s */\n" s
  end ;
  fprintf chan "/* barrier: %s */\n" (pp_barrier_loc arch) ;
  fprintf chan "/* launch: %s */\n" (pp_launch arch) ;
  fprintf chan "/* cache: %b */\n" (get_cache ()) ;
  fprintf chan "/* call: %b */\n" (get_call ()) ;
  fprintf chan "/* affinity: %s */\n" (pp_affinity (get_affinity ())) ;
  begin match get_logical_procs () with
  | None -> ()
  | Some xs ->
      fprintf chan "/* procs: {%s} */\n" (LexSplit.pp_ints xs)
  end ;
  fprintf chan "/* prealloc: %b */\n" (get_prealloc ()) ;
  fprintf chan "/* memory: %s */\n" (pp_memory arch) ;
  begin match get_stride () with
  | None -> ()
  | Some i ->   fprintf chan "/* stride: %i */\n" i
  end ;
  fprintf chan "/* safer: %s */\n" (pp_safer arch) ;
  fprintf chan "/* preload: %s */\n" (pp_preload (get_preload ())) ;
  fprintf chan "/* para: %s */\n" (pp_para ()) ;
  fprintf chan "/* speedcheck: %s */\n" (pp_speedcheck (get_speedcheck ())) ;
  begin match get_avail arch with
  | Some n -> fprintf chan "/* proc used: %i */\n" n
  | None -> ()
  end ;
  if get_isync () then begin
    fprintf chan "/* isync: true */\n"
  end ;
  close_quote chan ;
  ()

let report_machine arch chan =
(* Machine information *)
  if not (is_out ()) then begin
    output_string chan "Machine:"  ;
    exec "hostname" (output_line chan) ;
    begin match target_os arch with
    | Linux -> cat "/proc/cpuinfo" (output_line chan)
    | Mac -> exec "system_profiler SPHardwareDataType" (output_line chan)
    | AIX -> ()
    end
  end ;
(* Parameters of test *)
  report_parameters arch chan

module Make(Tar:Tar.S) (T:Test.S) =
  struct
    module A = T.A
    module C = T.C

(* Parallelism by shell script *)
    let do_shell = match get_para () with
    | Shell -> true
    | Self -> false



    let gcc () =
      assert (not (is_out ())) ;
      let gcc_opts = get_gcc_opts A.arch in
      let link_opts = get_link_opts A.arch in
      sprintf "%s %s%s"
        (get_gcc ())
        gcc_opts link_opts

    let gcc2as () =
      assert (not (is_out ())) ;
      let gcc_opts = get_gcc_opts A.arch in
      sprintf "%s %s" (get_gcc ()) gcc_opts 


    let compile_test name source =
      let sX = MyName.outname name ".exe" in
      if not (is_out ()) then begin
        let sX = Tar.outname sX in
        let source = Tar.outname source in
        let utils =
          let utils =
            match get_affinity () with
            | AffinityNone -> ["utils.c"; "outs.c"]
            | _ -> ["utils.c"; "outs.c"; "affinity.c"] in
          String.concat " " (List.map Tar.outname utils) in
        let com = sprintf "%s -o %s %s %s" (gcc()) sX utils source in
        exec_stdout com
      end ;
      sX

    let do_show out_chan in_chan =
      read_by_line in_chan
        (fun line prev ->
      let open Showcode in
      begin match prev with
      | Next -> fprintf out_chan "%s\n" line
      | No|Now -> ()
      end ;
      let here = Showcode.see line in
      begin match here with
      | Now -> fprintf out_chan "%s\n" line
      | No|Next -> ()
      end ;
      here)
    Showcode.No

    let showcode chan name =
      ignore (Misc.input_protect (do_show chan) name)

    let assemble_test chan name source =
      let sS = MyName.outname name ".s" in
      if is_out() then
        let sS = MyName.outname name ".t" in
        let com = sprintf "cat %s" sS in
        output_line chan com
      else begin
        let source = Tar.outname source in
        let sS = Tar.outname sS in
        let com = sprintf "%s -S -o %s %s"
            (gcc2as()) sS source in
        exec_stdout com ;
        showcode chan sS  ;
        output_line chan ""        
      end ;
      sS

    let do_dont = match Globals.get_speedcheck () with
    | AllSpeed -> true
    | NoSpeed|SomeSpeed -> false

    let run_test chan t name sX =
      let opts =
        sprintf "%s $LITMUSOPTS" (if !verbose > 0 then "-v" else "-q") in      
      if is_out() then begin
        let exe = Filename.concat "." sX in
        let com =
          match get_crossrun () with
          | RemoteNone -> sprintf "%s %s" exe opts
          | RemoteAdb ->  sprintf "dorun %s %s" exe opts
          | RemoteHost h ->
              let scp = sprintf "doscp %s" exe in
              let run = sprintf "%s %s" exe opts in
              let rm = sprintf "rm %s" exe in
              sprintf "%s && ssh -q -n -p $RPORT $RHOST \"%s && %s\""
                scp  run rm in
        let out_com chan com =
          if do_dont then
            fprintf chan "%s | tee $LOG\n" com
          else output_line chan com in
        begin match get_avail A.arch with
        | None -> out_com chan com
        | Some n ->
            let nprocs = List.length t.MiscParser.prog in
            let kmax = max 1 (n/nprocs) in
            if do_shell then begin
              if kmax > 1 then begin
                let tmp_file = sprintf "$TMPDIR/%i.log" in
                let output_com c =
                  output_string chan c ;
                  for k=1 to kmax do
                    fprintf chan " %s"  (tmp_file k)
                  done ;
                  output_line chan "" in
                for k=1 to kmax do
                  fprintf chan "%s -n 1 > %s &\n" com (tmp_file k)
                done ;
                output_line chan "wait" ;
                output_com "cat" ;
                output_com "rm -f"
              end else
                out_com chan (sprintf "%s -n 1" com)
            end else begin
              out_com chan com
            end
        end ;
        if do_dont then begin
          output_line chan
            "awk '/Sometimes/ {printf(\"touch %s.no\\n\",$2);}' $LOG | sh" ;
          output_line chan "rm -f $LOG"
        end
      end else begin
        let com = sprintf "%s %s" (Tar.outname sX) opts in
        exec_stdout com ;
        ()
      end 

    let dump_atom (loc,v) =
      sprintf "%s=%s" (A.pp_location loc) (A.V.pp_v v)

    let dump_constr chan = ConstrGen.dump_constraints chan dump_atom

    module D = SimpleDumper.Make
        (struct
           module A = A

          type state = A.state
          let dump_state chan st =
            Misc.pp_list chan " "
              (fun chan a ->
                output_string chan (dump_atom a) ;
                output_char chan ';')
              st

          type constr = C.constr
          let dump_constr = dump_constr

          type location = A.location
          let dump_location chan loc =
            output_string chan (A.pp_location loc)
        end)

open ConstrGen 

    let report name doc test source chan =
(* Static information *)
      let title = sprintf "%% Results for %s %%" name in
      let nice = String.make (String.length title) '%' in
      open_norun doc.Name.name chan ;
      open_quote chan ;
      output_line chan nice ;
      output_line chan title ;
      output_line chan nice ;
      D.dump chan doc test ;
(* Actual assembler *)
      output_string chan "Generated assembler\n" ;
      close_quote chan ;
      flush chan ;
(* End of static information *)
      let _sS = assemble_test chan name source in
(* Compile *)
      let sX = compile_test name source in
(* Run test *)
      flush chan ;
      run_test chan test name sX ;
      close_norun chan ;
      let sleep = Globals.get_sleep () in
      if sleep >= 0 then output_line chan "sleep $SLEEP" ;
      output_line chan "" ;
      flush chan

    let run name chan doc test source =
      report name doc test source chan

  end
